/**
 * OW2 Util
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: EjbJarDeployableMetadataFactory.java 4493 2009-02-02 09:25:06Z sauthieg $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.ee.metadata.ejbjar.impl;

import org.ow2.util.archive.api.ArchiveException;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.ee.deploy.api.deployable.EJBDeployable;
import org.ow2.util.ee.deploy.api.deployable.metadata.DeployableMetadataException;
import org.ow2.util.ee.metadata.ws.api.xml.struct.IWebservices;
import org.ow2.util.ee.metadata.ws.api.xml.merge.IClassMetadataFinder;
import org.ow2.util.ee.metadata.ws.api.IWebservicesCommonClassMetadata;
import org.ow2.util.ee.metadata.ws.impl.xml.parsing.Webservices12DeploymentDesc;
import org.ow2.util.ee.metadata.ws.impl.xml.merge.WebservicesMetadataMerger;
import org.ow2.util.ee.metadata.common.impl.AbsDeployableMetadataFactory;
import org.ow2.util.ee.metadata.common.impl.xml.parsing.ParsingException;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarClassMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarDeployableMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarDeployableMetadataFactory;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarFieldMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarMethodMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.struct.IApplicationException;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IAssemblyDescriptor;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IEnterpriseBeans;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IInterceptor;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IInterceptors;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.IMessageDriven;
import org.ow2.util.ee.metadata.ejbjar.api.xml.struct.ISession;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.EjbJarDeployableMetadataConfigurator;
import org.ow2.util.ee.metadata.ejbjar.impl.xml.helper.EjbJarMetadataMerge;
import org.ow2.util.ee.metadata.ejbjar.impl.xml.parsing.EJB3DeploymentDesc;
import org.ow2.util.ee.metadata.ejbjar.impl.xml.struct.EJB3;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.scan.api.IClassesLocator;
import org.ow2.util.scan.api.IScanner;
import org.ow2.util.scan.api.classlocator.ArchiveClassesLocator;
import org.ow2.util.scan.api.classlocator.ClassLoaderClassesLocator;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

/**
 * Factory to contruct ejb jar deployable metadata.
 * @param <E> an implementation of EJBDeployable
 * @param <D> an implementation of IEjbJarDeployableMetadata
 * @param <C> an implementation of IEjbJarClassMetadata
 * @param <M> an implementation of IEjbJarMethodMetadata
 * @param <F> an implementation of IEjbJarFieldMetadata
 * @author Gael Lalire
 */
public abstract class EjbJarDeployableMetadataFactory<
    E extends EJBDeployable<E>,
    D extends IEjbJarDeployableMetadata<E, D, C, M, F>,
    C extends IEjbJarClassMetadata<E, D, C, M, F>,
    M extends IEjbJarMethodMetadata<E, D, C, M, F>,
    F extends IEjbJarFieldMetadata<E, D, C, M, F>>
        extends AbsDeployableMetadataFactory implements IEjbJarDeployableMetadataFactory<E, D, C, M, F> {

    /**
     * Logger.
     */
    private static final Log LOGGER = LogFactory.getLog(EjbJarDeployableMetadataFactory.class);

    /**
     * Default Constructor.
     */
    public EjbJarDeployableMetadataFactory() {
    }

    /**
     * Constructor a scanner.
     * @param scanner the scanner
     */
    public EjbJarDeployableMetadataFactory(final IScanner scanner) {
        super(scanner);
    }

    /**
     * @param ejbDeployable an ejb deployable
     * @return metadata constructed using ejb3Deployable
     * @throws DeployableMetadataException if metadata can not be created
     */
    public D createDeployableMetadata(final E ejbDeployable) throws DeployableMetadataException {
        return createDeployableMetadata(ejbDeployable, (List<? extends IClassesLocator>) null);
    }

    /**
     * @param deployable an ejb3 deployable
     * @param classLoader a classloader to use if a class is not in archive
     * @return metadata constructed using ejb3Deployable
     * @throws DeployableMetadataException if metadata can not be created
     */
    public D createDeployableMetadata(final E deployable, final ClassLoader classLoader)
            throws DeployableMetadataException {
        List<? extends IClassesLocator> classesLocatorList = null;
        if (classLoader != null) {
            classesLocatorList = Arrays.asList(new ClassLoaderClassesLocator(classLoader));
        } else {
            classesLocatorList = Arrays.asList(new ClassLoaderClassesLocator(ClassLoader.getSystemClassLoader()));
        }
        return createDeployableMetadata(deployable, classesLocatorList);
    }

    /**
     * Extract a class list from xml.
     * @param ejb3 representation of ejb-jar.xml
     * @param classNameSet a set to fill with classes name referenced in ejb-jar.xml
     */
    private void completeClassList(final EJB3 ejb3, final Set<String> classNameSet) {
        if (ejb3 != null) {

            // Get enterprise beans classes
            IEnterpriseBeans enterpriseBeans = ejb3.getEnterpriseBeans();
            if (enterpriseBeans != null) {
                // get session beans
                List<ISession> sessionList = enterpriseBeans.getSessionList();
                for (ISession session : sessionList) {
                    // get ejb-class
                    addClassToSetIfNotNull(session.getEjbClass(), classNameSet);
                }
                List<IMessageDriven> messageDrivenList = enterpriseBeans.getMessageDrivenList();
                for (IMessageDriven messageDriven : messageDrivenList) {
                    // get ejb-class
                    addClassToSetIfNotNull(messageDriven.getEjbClass(), classNameSet);
                }
            }

            // Get interceptor classes
            IInterceptors interceptors = ejb3.getInterceptors();
            if (interceptors != null) {
                List<IInterceptor> interceptorList = interceptors.getInterceptorList();
                if (interceptorList != null) {
                    for (IInterceptor interceptor : interceptorList) {
                        addClassToSetIfNotNull(interceptor.getInterceptorClass(), classNameSet);
                    }
                }
            }

            // Get application exception classes
            IAssemblyDescriptor assemblyDescriptor = ejb3.getAssemblyDescriptor();
            if (assemblyDescriptor != null) {
                List<IApplicationException> applicationExceptionList = assemblyDescriptor.getApplicationExceptionList();
                if (applicationExceptionList != null) {
                    for (IApplicationException applicationException : applicationExceptionList) {
                        addClassToSetIfNotNull(applicationException.getClassName(), classNameSet);
                    }
                }
            }

        }

    }

    /**
     * Build a new deployable metadata configurator for the given ddeployable.
     * @param ejbDeployable the given deployable
     * @return an instance of a metadataconfigurator
     */
    protected abstract EjbJarDeployableMetadataConfigurator<E, D, C, M, F> createEjbJarDeployableMetadataConfigurator(
            final E ejbDeployable);

    /**
     * Build a new DeployableMetadata object from the given deployable.
     * @param ejbDeployable the given deployable
     * @return a new metadata
     */
    protected abstract D createEjbJaDeployableMetadata(final E ejbDeployable);

    /**
     * @param ejb3Deployable an ejb3 deployable
     * @param ejbJarExternLibraries extern libraries
     * @return metadata constructed using ejb3Deployable
     * @throws DeployableMetadataException if metadata can not be created
     */
    public D createDeployableMetadata(final E ejb3Deployable,
            final List<? extends IClassesLocator> ejbJarExternLibraries)
            throws DeployableMetadataException {

        // for time debugging
        long tAnalyzeStart = System.currentTimeMillis();

        IArchive archive = ejb3Deployable.getArchive();

        // Get EJB3 DD.
        EJB3 ejb3;
        try {
            ejb3 = EJB3DeploymentDesc.getEjb3(archive);
        } catch (ParsingException e) {
            throw new DeployableMetadataException(e);
        }

        // Get Webservices DD.
        IWebservices webservices = null;
        try {
            webservices = Webservices12DeploymentDesc.getWebservices12(archive,
                    Webservices12DeploymentDesc.Type.EJBJAR);
        } catch (ParsingException e) {
            throw new DeployableMetadataException(e);
        }

        EjbJarDeployableMetadataConfigurator<E, D, C, M, F> ejbJarArchiveMetadataConfigurator
            = createEjbJarDeployableMetadataConfigurator(ejb3Deployable);

        Set<String> classNameSet = new HashSet<String>();

        beforeScan(ejbJarArchiveMetadataConfigurator.getEjbJarArchiveMetadata());

        // if metadata is complete then filter class to scan
        if (ejb3 != null && ejb3.isMetadataComplete()) {
            // disable annotation parsing
            ejbJarArchiveMetadataConfigurator.setAnnotationParsingDesactived(true);

        } else {
            // enable annotation parsing because metadata descriptor is not
            // complete
            ejbJarArchiveMetadataConfigurator.setAnnotationParsingDesactived(false);

            Iterator<String> entryIterator;
            try {
                entryIterator = archive.getEntries();
            } catch (ArchiveException e) {
                throw new DeployableMetadataException(e);
            }
            while (entryIterator.hasNext()) {
                String entry = entryIterator.next();
                if (entry.endsWith(CLASS_SUFFIX)) {
                    classNameSet.add(entry.substring(0, entry.length() - CLASS_SUFFIX.length()));
                }
            }
        }

        // get class to scan from deployment descriptor
        completeClassList(ejb3, classNameSet);

        // do annotation scan
        scan(classNameSet, new ArchiveClassesLocator(archive), ejbJarExternLibraries, ejbJarArchiveMetadataConfigurator);

        // get scan result
        final D ejbJarArchiveMetadata = ejbJarArchiveMetadataConfigurator.getEjbJarArchiveMetadata();

        // prepare the metadata for merge
        ejbJarArchiveMetadata.setEjb3(ejb3);
        ejbJarArchiveMetadata.setWebservices12(webservices);

        // Call after scan
        afterScan(ejbJarArchiveMetadata);

        // Merge DD with annotations metadata
        EjbJarMetadataMerge.merge(ejbJarArchiveMetadata);

        WebservicesMetadataMerger.merge(ejbJarArchiveMetadata, new IClassMetadataFinder() {

            public IWebservicesCommonClassMetadata findClassMetadata(String link) {
                return ejbJarArchiveMetadata.getEjbJarClassMetadataForEjbName(link);
            }
        });

        afterMerge(ejbJarArchiveMetadata);

        // time if debugging
        if (LOGGER.isDebugEnabled()) {
            long tAnalyzeStartEnd = System.currentTimeMillis();
            LOGGER.debug("Analyze of file {0} took {1} ms.", archive.getName(), (tAnalyzeStartEnd - tAnalyzeStart));
        }

        return ejbJarArchiveMetadata;
    }

    /**
     * @return the class returned by creation of metadata :
     *         IEJB3DeployableMetadata.class
     */
    @SuppressWarnings("unchecked")
    public Class<D> getDeployableMetadataClass() {
        return (Class<D>) (Class<?>) IEjbJarDeployableMetadata.class;
    }

    /**
     * This method is called before metadata is filled.
     * @param ejbJarArchiveMetadata the given archive metadata.
     */
    public void beforeScan(final D ejbJarArchiveMetadata) {

    }

    /**
     * This method is called when metadata is filled.
     * @param ejbJarArchiveMetadata the given archive metadata.
     */
    public void afterScan(final D ejbJarArchiveMetadata) {

    }

    /**
     * This method is called when metadata is merged.
     * @param ejbJarArchiveMetadata the given archive metadata.
     */
    public void afterMerge(final D ejbJarArchiveMetadata) {

    }

}
